@@ -6,6 +6,7 @@
 
 #include <stdio.h>
 #include <stdlib.h>
+#include <string.h> // memcpy
 #include "swf_define.h"
 
 #ifdef HAVE_PNG
@@ -87,7 +88,7 @@ void *
 pngconv_png2lossless(unsigned char *png_data, unsigned long png_data_len,
                      int *tag_no, int *format,
                      unsigned short *width, unsigned short *height,
-                     void **colormap, int *colormap_count) {
+                     void **colormap, int *colormap_count, int rgb15) {
     volatile png_structp png_ptr = NULL;
     volatile png_infop png_info_ptr = NULL;
     my_png_buffer png_buff;
@@ -145,6 +146,11 @@ pngconv_png2lossless(unsigned char *png_data, unsigned long png_data_len,
                  NULL, NULL, NULL);
     *width = (unsigned short) png_width;
     *height = (unsigned short) png_height;
+
+    if ((rgb15 > 0) && (color_type !=  PNG_COLOR_TYPE_RGB)) { // warning
+        fprintf(stderr, "rgb15 is %d but color_type is not PNG24\n", rgb15);
+    }
+
     switch(color_type) {
     case PNG_COLOR_TYPE_PALETTE:
         *format = 3;
@@ -157,7 +163,11 @@ pngconv_png2lossless(unsigned char *png_data, unsigned long png_data_len,
         }
         break;
     case PNG_COLOR_TYPE_RGB:
-        *format = 5;
+        if (rgb15 > 0) {
+            *format = 4;
+        } else {
+            *format = 5;
+        }
         *tag_no = 20; /* DefineBitsLossless */
         break;
     case PNG_COLOR_TYPE_RGB_ALPHA:
@@ -225,7 +235,10 @@ pngconv_png2lossless(unsigned char *png_data, unsigned long png_data_len,
                 free(*colormap);
                 *colormap = NULL;
                 free(indices_data);
-                // free png image datas
+                for (y=0 ; y < png_height ; y++) {
+                    free(png_image_data[y]);
+                }
+                free(png_image_data);
                 return NULL;
             }
             bitstream_input(bs, png_image_data[y], png_width);
@@ -363,12 +376,31 @@ pngconv_lossless2png(void *image_data,
             }
         } else {
             swf_rgba_t *rgba_list  = index_data;
+            png_bytep trans = NULL;
+            int num_trans = 0;
+            png_color_16p trans_values = NULL;
             for (i=0 ; i < index_data_count ; i++) {
                 png_palette[i].red   = rgba_list[i].red;
                 png_palette[i].green = rgba_list[i].green;
                 png_palette[i].blue  = rgba_list[i].blue;
-//                png_palette[i].alpha = rgba_list[i].alpha;
             }
+            // scanning to end of transparent pixel
+            for (i = index_data_count - 1 ; 0 <= i ; i--) {
+                if (rgba_list[i].alpha < 254) { // 254 XXX
+                    break;
+                }
+            }
+            num_trans = i + 1;
+            if (num_trans > 0) {
+                trans = malloc(num_trans);
+                for (i = 0 ; i < num_trans ; i++) {
+                    trans[i] = rgba_list[i].alpha;
+                }
+                png_set_tRNS(png_ptr, png_info_ptr, trans, num_trans,
+                             trans_values);
+                free(trans);
+            }
+
         }
         png_set_PLTE( png_ptr, png_info_ptr, png_palette, index_data_count);
         free(png_palette);
