@@ -227,7 +227,7 @@ swf_tag_shape_bitmap_get_refcid_list(swf_tag_t *tag, int *cid_list_num) {
                 if (fill_style->bitmap.bitmap_ref != 0xffff) {
                     if (cid_list_alloc <= *cid_list_num) {
                         cid_list_alloc *= 2;
-                        tmp = realloc(cid_list, cid_list_alloc);
+                        tmp = realloc(cid_list, sizeof(int) * cid_list_alloc);
                         if (tmp == NULL) {
                             fprintf(stderr, "swf_tag_shape_bitmap_get_refcid_list: Can't realloc memory (%p, %d)\n", cid_list, cid_list_alloc);
                             free(cid_list);
@@ -368,6 +368,7 @@ swf_tag_shape_output_detail(swf_tag_t *tag, unsigned long *length,
     long offset_of_offset_morph = 0;
     long tmp_offset_byte = 0;
     long tmp_offset_bit = 0;
+    (void) swf;
     *length = 0;
     // build context
     swf_tag_shape->_current_fill_style_num = 0;
@@ -489,14 +490,17 @@ swf_tag_shape_destroy_detail(swf_tag_t *tag) {
 }
 
 int
-swf_tag_shape_apply_matrix_factor(void *detail, int shape_id,
+swf_tag_shape_apply_matrix_factor(void *detail, int shape_id, int bitmap_id,
                                   double scale_x, double scale_y,
                                   double rotate_rad,
                                   signed int trans_x,
                                   signed int trans_y) {
     int i;
     swf_tag_shape_detail_t *swf_tag_shape = (swf_tag_shape_detail_t *) detail;
+    swf_fill_style_array_t *fill_styles;
     swf_fill_style_t *fill_style;
+    swf_shape_record_t *shape_record = NULL;
+
     if (detail == NULL) {
         fprintf(stderr, "swf_tag_shape_apply_matrix_factor: detail == NULL\n");
         return 1;
@@ -504,34 +508,70 @@ swf_tag_shape_apply_matrix_factor(void *detail, int shape_id,
     if (shape_id != swf_tag_shape->shape_id) {
         return 1;
     }
-    for (i = 0 ; i < swf_tag_shape->shape_with_style.styles.fill_styles.count ; i++) {
-        fill_style = &(swf_tag_shape->shape_with_style.styles.fill_styles.fill_style[i]);
-        switch (fill_style->type) {
-          case 0x10: // linear gradient fill
-          case 0x12: // radial gradient fill
-          case 0x13: // focal  gradient fill
-              swf_matrix_apply_factor(&(fill_style->gradient.gradient_matrix), scale_x, scale_y,
-                                      rotate_rad, trans_x, trans_y);
-              break;
-          case 0x40: // tilled  bitmap fill with smoothed edges
-          case 0x41: // clipped bitmap fill with smoothed edges
-          case 0x42: // tilled  bitmap fill with hard edges
-          case 0x43: // clipped bitmap fill with hard edges
-              swf_matrix_apply_factor(&(fill_style->bitmap.bitmap_matrix),
-                                      scale_x, scale_y, rotate_rad,
-                                      trans_x, trans_y);
-              break;
-          default:
-            fprintf(stderr, "swf_tag_shape_apply_matrix_factor: unknown fill_style->type=%d\n",
-                    fill_style->type);
-            return 1;
+    fill_styles = &(swf_tag_shape->shape_with_style.styles.fill_styles);
+
+    if (! swf_tag_shape->is_morph) {
+        shape_record = &(swf_tag_shape->shape_with_style.shape_records);
+    }
+
+    while (fill_styles) {
+        for (i = 0 ; i < fill_styles->count ; i++) {
+            fill_style = &(fill_styles->fill_style[i]);
+            switch (fill_style->type) {
+            case 0x10: // linear gradient fill
+            case 0x12: // radial gradient fill
+            case 0x13: // focal  gradient fill
+                if (bitmap_id < 0) {
+                    swf_matrix_apply_factor(&(fill_style->gradient.gradient_matrix), scale_x, scale_y,
+                                            rotate_rad, trans_x, trans_y);
+                }
+                break;
+            case 0x40: // tilled  bitmap fill with smoothed edges
+            case 0x41: // clipped bitmap fill with smoothed edges
+            case 0x42: // tilled  bitmap fill with hard edges
+            case 0x43: // clipped bitmap fill with hard edges
+                if ((bitmap_id < 0) ||
+                    (bitmap_id == fill_style->bitmap.bitmap_ref)) {
+                    swf_matrix_apply_factor(&(fill_style->bitmap.bitmap_matrix),
+                                            scale_x, scale_y, rotate_rad,
+                                            trans_x, trans_y);
+                }
+                break;
+            case 0x00: // solid
+                // nothing to do
+                break;
+            default:
+                fprintf(stderr, "swf_tag_shape_apply_matrix_factor: unknown fill_style->type=%d\n",
+                        fill_style->type);
+                return 1;
+            }
+        }
+        fill_styles = NULL;
+        while (shape_record) {
+            int first_bit = (shape_record->first_6bits >> 5) & 1;
+            int next_5bits = shape_record->first_6bits & 0x1f;
+            if (first_bit == 0) {
+                if (next_5bits == 0) { // end record
+                    shape_record = NULL;
+                    break; // end
+                } else { // setup record
+                    if (shape_record->shape.shape_setup.shape_has_new_styles) {
+                        fill_styles = &(shape_record->shape.shape_setup.styles.fill_styles);
+                        shape_record = shape_record->next;
+                        break; // next loop
+                    }
+                    shape_record = shape_record->next;
+                }
+            } else { // edge record
+                shape_record = shape_record->next;
+            }
         }
     }
     return 0;
 }
 
 int
-swf_tag_shape_apply_rect_factor(void *detail, int shape_id,
+swf_tag_shape_apply_rect_factor(void *detail, int shape_id, int bitmap_id,
                                 double scale_x, double scale_y,
                                 signed int trans_x,
                                 signed int trans_y) {
@@ -544,18 +584,22 @@ swf_tag_shape_apply_rect_factor(void *detail, int shape_id,
     if (shape_id != swf_tag_shape->shape_id) {
         return 1;
     }
+    // XXX: fix me.
     swf_rect_apply_factor(&(swf_tag_shape->rect),
                           scale_x, scale_y, trans_x, trans_y);
-    
+    // TODO: bitmap_ref check
     swf_shape_record_edge_apply_factor(&(swf_tag_shape->shape_with_style.shape_records),
                                        scale_x, scale_y, trans_x, trans_y);
     return 0;
 }
 
 int
-swf_tag_shape_apply_type_tilled(void *detail, int shape_id) {
-    int i, count;
+swf_tag_shape_apply_type_tilled(void *detail, int shape_id, int bitmap_id) {
+    int i;
     swf_tag_shape_detail_t *swf_tag_shape = (swf_tag_shape_detail_t *) detail;
+    swf_fill_style_array_t *fill_styles;
+    swf_shape_record_t *shape_record = NULL;
+
     if (detail == NULL) {
         fprintf(stderr, "swf_tag_shape_apply_type_tilled: detail == NULL\n");
         return 1;
@@ -565,16 +609,48 @@ swf_tag_shape_apply_type_tilled(void *detail, int shape_id) {
 	      shape_id, swf_tag_shape->shape_id);
         return 1;
     }
-    count = swf_tag_shape->shape_with_style.styles.fill_styles.count;
-    for (i = 0 ; i < count ; i++) {
-        swf_fill_style_t *fill_style = &(swf_tag_shape->shape_with_style.styles.fill_styles.fill_style[i]);
-        switch (fill_style->type) {
-          case 0x41: // clipped bitmap fill with smoothed edges
-            fill_style->type = 0x40; // tilled  bitmap fill with smoothed edges
-            break;
-          case 0x43: // clipped bitmap fill with hard edges
-            fill_style->type = 0x42; // tilled  bitmap fill with hard edges
-            break;
+
+    fill_styles = &(swf_tag_shape->shape_with_style.styles.fill_styles);
+
+    if (! swf_tag_shape->is_morph) {
+        shape_record = &(swf_tag_shape->shape_with_style.shape_records);
+    }
+
+    while (fill_styles) {
+        for (i = 0 ; i < fill_styles->count; i++) {
+            swf_fill_style_t *fill_style = &(fill_styles->fill_style[i]);
+            switch (fill_style->type) {
+            case 0x41: // clipped bitmap fill with smoothed edges
+                if ((bitmap_id < 0) || (bitmap_id == fill_style->bitmap.bitmap_ref)) {
+                    fill_style->type = 0x40; // tilled  bitmap fill with smoothed edges
+                }
+                break;
+            case 0x43: // clipped bitmap fill with hard edges
+                if ((bitmap_id < 0) || (bitmap_id == fill_style->bitmap.bitmap_ref)) {
+                    fill_style->type = 0x42; // tilled  bitmap fill with hard edges
+                }
+                break;
+            }
+        }
+        fill_styles = NULL;
+        while (shape_record) {
+            int first_bit = (shape_record->first_6bits >> 5) & 1;
+            int next_5bits = shape_record->first_6bits & 0x1f;
+            if (first_bit == 0) {
+                if (next_5bits == 0) { // end record
+                    shape_record = NULL;
+                    break; // end
+                } else { // setup record
+                    if (shape_record->shape.shape_setup.shape_has_new_styles) {
+                        fill_styles = &(shape_record->shape.shape_setup.styles.fill_styles);
+                        shape_record = shape_record->next;
+                        break; // next loop
+                    }
+                    shape_record = shape_record->next;
+                }
+            } else { // edge record
+                shape_record = shape_record->next;
+            }
         }
     }
     return 0;
